@@ -957,15 +957,42 @@ esc_subsystem_info:
 	 */
 	string SysFSTreeCollector::getDevTreePath(string sysPath)
 	{
+		string sysFsPath, procDtPath, firmwareDtBase;
 		struct stat astats;
 		const char *buf;
-		char buf2[512];
+		char buf2[PATH_MAX];
 		FILE *fi;
 
 		HelperFunctions::fs_fixPath(sysPath);
-		sysPath += "/devspec";
+		sysFsPath = sysPath + "/devspec";
+		procDtPath = sysPath + "/of_node";
+		firmwareDtBase = "/sys/firmware/devicetree/base";
 
-		buf = sysPath.c_str();
+		/*
+		 * Check for existence of of_node symlink and return the path it
+		 * points to.  devspec is obsolete now, most of the devices
+		 * populate of_node and devspec too. Prefer of_node over
+		 * devspec, where it exist and fall back to older logic, in case
+		 * of of_node not populated.
+		 */
+		buf = procDtPath.c_str();
+		if ((lstat(buf, &astats)) == 0) {
+			// of_node is symlink, follow the link
+			if ( realpath( buf, buf2 ) == NULL ) {
+				return string ("");
+			}
+
+			/*
+			 * Trim the leading "/sys/firmware/devicetree/base" from the real
+			 * path to match the assumption from devspec format, while used
+			 * with /proc/devicetree.
+			 */
+			procDtPath = string(buf2);
+			procDtPath.replace(0, firmwareDtBase.length(), "");
+			return procDtPath;
+		}
+
+		buf = sysFsPath.c_str();
 		if ((lstat(buf, &astats)) != 0) {
 			return string("");
 		}
@@ -985,7 +1012,10 @@ esc_subsystem_info:
 
 		fclose(fi);
 		// cleanup
-		return string(buf2);
+		if (strcmp(buf2, "(null)"))
+			return string(buf2);
+
+		return string("");
 	}
 
 	/* Collects devices on system, then returns how many were found.
@@ -1377,6 +1407,22 @@ ERROR:
 		return 0;
 	}
 
+	void SysFSTreeCollector::fillPciNvmeVpd( Component* fillMe )
+	{
+		int device_fd;
+		struct stat myDir;
+		string path;
+		path = fillMe->sysFsNode.getValue() + "/nvme";
+		if (stat(path.c_str(), &myDir) < 0)
+			return;
+		device_fd = device_open(fillMe);
+		if (device_fd < 0)
+			return;
+		collectNvmeVpd(fillMe, device_fd);
+		close(device_fd);
+		return;
+	}
+
 	/* Parse VPD file */
 	void SysFSTreeCollector::fillPciDevVpd( Component* fillMe )
 	{
@@ -1434,7 +1480,8 @@ ERROR:
 		if( mPciTable != NULL )
 		{
 			// Fill Manufacturer Name
-			if( subMan == UNKNOWN_ID )
+			if( subMan == UNKNOWN_ID ||
+			    (mPciTable->getName( subMan ) == "Unknown") )
 			{
 				if( manID != UNKNOWN_ID )
 				{
@@ -1489,6 +1536,9 @@ ERROR:
 		/* Fill PCI device VPD info */
 		fillPciDevVpd(fillMe);
 
+		/* Fill NVME device VPD info using f1h log page */
+		fillPciNvmeVpd(fillMe);
+
 		// Read the pci config file for Device Specific (YC)
 		os.str( "" );
 		os << fillMe->sysFsNode.dataValue << "/config";
@@ -1669,6 +1719,9 @@ ERROR:
 		 */
 		newDevDir = dev_syspath + "/" + dev_childname;
 		str = getAttrValue( newDevDir, "dev" );
+		if (str.empty())
+			return;
+
 		beg = end = 0;
 		while (end < (int) str.length() && str[end] != ':')
 			end++;
@@ -1784,6 +1837,10 @@ ERROR:
 
 	void SysFSTreeCollector::fillFirmware( Component* fillMe )
 	{
+		string result = "";
+		string path = fillMe->getID();
+		const string firmwareAttributes[] = {"fw_version", "firmware_rev"};
+		const size_t firmwareAttributesSize = sizeof(firmwareAttributes) / sizeof(firmwareAttributes[0]);
 		string classNode = fillMe->getClassNode();
 		if (classNode.length() > 0) {
 			fillMe->mFirmwareVersion.setValue( getAttrValue( classNode,
@@ -1791,6 +1848,26 @@ ERROR:
 
 			fillMe->mFirmwareLevel.setValue( getAttrValue( classNode,
 					"fwrev" ), 30, __FILE__, __LINE__ );
+
+			fillMe->mFirmwareVersion.setValue( getAttrValue( classNode,
+					"firmware_rev" ), 30, __FILE__, __LINE__ );
+		}
+
+		if (fillMe->mFirmwareLevel.dataValue.empty()) {
+			result = searchFile(path, "fwrev");
+			if (!result.empty()) {
+				fillMe->mFirmwareLevel.setValue( getAttrValue( result,
+							"fwrev" ), 30, __FILE__, __LINE__ );
+			}
+		}
+
+		for (size_t i = 0; i < firmwareAttributesSize && fillMe->mFirmwareVersion.dataValue.empty(); i++) {
+			result = searchFile(path, firmwareAttributes[i]);
+			if (!result.empty()) {
+				fillMe->mFirmwareVersion.setValue( getAttrValue( result,
+							firmwareAttributes[i]), 30, __FILE__, __LINE__ );
+				break;
+			}
 		}
 	}
 
